"""
Reactive Reality Machine Learning Config System - Configuration object
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from decimal import Context
import functools
from functools import partial
import io
import logging
import re
import sys
from collections.abc import Mapping
from enum import Enum
from numbers import Real
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

if TYPE_CHECKING:
    from .config import Configuration

YAECS_LOGGER = logging.getLogger(__name__)
ConfigDeclarator = Union[str, dict]
ConfigInput = Union[List[ConfigDeclarator], ConfigDeclarator]
Hooks = Union[Dict[str, List[str]], List[str]]
ProcessingFunction = Union[Callable[[Any], Any], str]
ProcessingOrder = Union[Real, 'Priority']
ProcessingFunctions = Union[ProcessingFunction, Tuple[Union[ProcessingFunction, ProcessingOrder]]]
TypeHint = Union[type, tuple, list, dict, set, int]
VariationDeclarator = Union[List[ConfigDeclarator], Dict[str, ConfigDeclarator]]
YAML_EXPRESSIONS = {
    "null": re.compile(r'''^(?: ~
                    |null|Null|NULL
                    | )$''', re.X),
    "bool": re.compile(r'''^(?:yes|Yes|YES|no|No|NO
                    |true|True|TRUE|false|False|FALSE
                    |on|On|ON|off|Off|OFF)$''', re.X),
    "int": re.compile(r'''^(?:[-+]?0b[0-1_]+
                    |[-+]?0[0-7_]+
                    |[-+]?(?:0|[1-9][0-9_]*)
                    |[-+]?0x[0-9a-fA-F_]+
                    |[-+]?[1-9][0-9_]*(?::[0-5]?[0-9])+)$''', re.X),
    "float": re.compile(r'''^(?:[-+]?(?:[0-9][0-9_]*)\.[0-9_]*(?:[eE][-+][0-9]+)?
                    |\.[0-9][0-9_]*(?:[eE][-+][0-9]+)?
                    |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\.[0-9_]*
                    |[-+]?\.(?:inf|Inf|INF)
                    |\.(?:nan|NaN|NAN))$''', re.X)
}
TYPE_HINT_MAPPING_STARTS = {"tuple_0": "(", "tuple_1": "union[", "nonetuple": "optional[",
                            "list_0": "[", "list_1": "list[",
                            "set_0": "d", "set_1": "dict["}
TYPE_HINT_MAPPING_ENDS = {"tuple_0": ")", "tuple_1": "]", "nonetuple": "]",
                          "list_0": "]", "list_1": "]",
                          "set_0": "/d", "set_1": "]"}
TYPE_HINT_SIMPLE_TYPES = {"none": None, "int": int, "float": float, "bool": bool, "str": str, "list": list,
                          "dict": dict, "any": 0}


class NoValue:
    """ Used to represent a default value not modified by the user. """


class Priority(Enum):
    """ Define priority levels which can be used to qualify when a processing function should be performed. """
    ALWAYS_FIRST = -20
    OFTEN_FIRST = -10
    INDIFFERENT = 0
    SITUATIONAL = 0
    OFTEN_LAST = 10
    ALWAYS_LAST = 20

    def __hash__(self):
        return hash(self.value)

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        if isinstance(other, Real):
            return self.value > other
        if isinstance(other, str):
            return self.value > getattr(self.__class__, other)
        return NotImplemented

    def __rgt__(self, other):
        return self < other

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        if isinstance(other, Real):
            return self.value < other
        if isinstance(other, str):
            return self.value < getattr(self.__class__, other)
        return NotImplemented

    def __rlt__(self, other):
        return self > other

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        if isinstance(other, Real):
            return self.value >= other
        if isinstance(other, str):
            return self.value >= getattr(self.__class__, other)
        return NotImplemented

    def __rge__(self, other):
        return self <= other

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        if isinstance(other, Real):
            return self.value <= other
        if isinstance(other, str):
            return self.value <= getattr(self.__class__, other)
        return NotImplemented

    def __rle__(self, other):
        return self >= other

    def __eq__(self, other):
        if self.__class__ is other.__class__:
            return self.value == other.value
        if isinstance(other, Real):
            return self.value == other
        if isinstance(other, str):
            return self.name == other
        return NotImplemented

    def __req__(self, other):
        return self == other


class TqdmLogFormatter:
    """
    Context setting formatters used in logging handlers for tqdm bars. See https://github.com/tqdm/tqdm/issues/313
    """

    def __init__(self, logger):
        self._logger = logger
        self.__original_formatters = None

    def __enter__(self):
        self.__original_formatters = list()

        for handler in self._logger.handlers:
            self.__original_formatters.append(handler.formatter)

            handler.terminator = ''
            formatter = logging.Formatter('%(message)s')
            handler.setFormatter(formatter)

        return self._logger

    def __exit__(self, exc_type, exc_value, exc_traceback):
        for handler, formatter in zip(self._logger.handlers, self.__original_formatters):
            handler.terminator = '\n'
            handler.setFormatter(formatter)


class TqdmLogger(io.StringIO):
    """File to use in tqdm to make it log its bars to a logger. See https://github.com/tqdm/tqdm/issues/313"""

    def __init__(self, logger):
        super().__init__()

        self._logger = logger

    def write(self, buffer):
        with TqdmLogFormatter(self._logger) as logger:
            logger.info(buffer)

    def flush(self):
        pass


def assign_order(order: ProcessingOrder = Priority.INDIFFERENT) -> Callable[[Callable], Callable]:
    """
    Decorator used to give an order to a processing function. If several processing functions would be called at a given
    step, they are called in increasing order.

    :param order: order to give the function
    :return: decorated function
    """
    def decorator_order(func: Callable) -> Callable:
        if not hasattr(func, "yaecs_metadata"):
            set_function_attribute(func, "yaecs_metadata", {})
        if "name" not in func.yaecs_metadata:
            func.yaecs_metadata["name"] = func.__name__
        func.yaecs_metadata["order"] = order
        return func

    return decorator_order


def assign_yaml_tag(processor_tag: str, processor_type: str,
                    replacement_type_hint: str = "Any") -> Callable[[Callable], Callable]:
    """
    Decorator used to mark a function as a processor added automatically as pre or post processing function (as
    defined by processor_type) to parameters tagged with !<processor_tag>. Their type hint will be replaced by
    the type hint defined as replacement_type_hint if this is the first processing function to be called on the
    parameter.

    :param processor_tag: tag to use to mark a param in YAML as auto-processed by this function
    :param processor_type: 'pre' or 'post', type of processing function to add
    :param replacement_type_hint: type hint to use for any param tagged with this auto-processor
    :return: decorated function
    """
    def decorator_tag_assignment(func: Callable) -> Callable:
        if "yaecs_metadata" not in func.__dict__:
            func.__dict__["yaecs_metadata"] = {}
        func.__dict__["yaecs_metadata"].update({
            "tag": processor_tag,
            "name": func.__name__,
            "processing_type": processor_type,
            "input_type": replacement_type_hint,
        })
        return func

    return decorator_tag_assignment


def check_type(type_or_types: TypeHint, name: Optional[str] = None) -> Callable:
    """
    Returns a processing function that checks for given type. Can be used for example with the following line in a
    parameters post-processing dict:
    "parameter_that_should_be_int": check_type(int)

    * The type can be any of None, bool, int, float, str, dict, list. The value 0 instead means no type check.
    * Unions are denoted by tuples of types.
    * You can specify the type of the elements of your lists by using a list of types. This list should contain
        either one type (in which case the list is expected to only contain elements of that type) or as many types as
        there are elements in the list (in which case each element is tested with the corresponding type)
    * You can specify the type of the elements of your dicts by using a dict or a set of types. If you use a set, it
        can only contain one type (in which case the dict is expected to contain only values of that type).
        If you use a dict of types, the keys used in that dict that match the keys in the parameter will be checked
        using the values as types.

    :param type_or_types: type for which to create the function
    :param name: name of the parameter to check
    :return: the processing function
    """
    def _check_type(value: Any, type_to_check: TypeHint, original_type: TypeHint, name: str) -> Any:
        def _wrong_type() -> None:
            is_full = original_type == type_to_check
            if name is None:
                header = f"{'Value' if is_full else 'Part of value'} '{value}'"
            else:
                header = f"{'Parameter' if is_full else 'Part of parameter'} '{name}' (value : {value})"
            checked_type = type(type_to_check) if isinstance(type_to_check, (list, dict, set)) else type_to_check
            raise ValueError(f"{header} has incorrect type '{type(value)}'. Expected '{checked_type}'.")

        if isinstance(type_to_check, tuple):
            if not type_to_check:
                raise ValueError("Undefined behaviour for empty tuples. Maybe you meant to use an empty list or dict ?")
            fails = True
            for to_check in type_to_check:
                try:
                    _check_type(value, to_check, original_type, name)
                except ValueError:
                    pass
                else:
                    fails = False
            if fails:
                _wrong_type()

        elif isinstance(type_to_check, list):
            if not isinstance(value, list):
                _wrong_type()
            if len(type_to_check) > 1:
                if len(type_to_check) != len(value):
                    raise ValueError("When providing a list of types, its length must be one or match the length of"
                                     " the value.")
                for v_to_check, t_to_check in zip(value, type_to_check):
                    _check_type(v_to_check, t_to_check, original_type, name)
            else:
                types = type_to_check[0] if type_to_check else 0
                for i in value:
                    _check_type(i, types, original_type, name)

        elif isinstance(type_to_check, dict):
            if not isinstance(value, dict):
                _wrong_type()
            if not type_to_check:
                raise ValueError("Undefined behaviour for empty dicts. Maybe you meant to use an empty list or "
                                 "{\"type\": ...} ?")
            if len(type_to_check) > 1:
                raise ValueError("When providing a dict of types, its length must be 1. Maybe you meant to use a"
                                 " tuple ?")
            for i in value:
                _check_type(value[i], type_to_check[list(type_to_check.keys())[0]], original_type, name)

        elif type_to_check != 0 and type_to_check is not None and not isinstance(value, type_to_check):
            if not (type_to_check is float and isinstance(value, int)):
                _wrong_type()

        elif type_to_check is None and value is not None:
            _wrong_type()
        return value

    return partial(_check_type, type_to_check=type_or_types, original_type=type_or_types, name=name)


def compare_string_pattern(name: str, pattern: str) -> bool:
    """
    Returns True when string 'name' matches string 'pattern', with the '*' character matching any number of characters.

    :param name: name to compare
    :param pattern: pattern to match
    :return: result of comparison
    """
    pattern = pattern.strip(" ").split("*")
    if len(pattern) == 1:
        return pattern[0] == name
    if not (name.startswith(pattern[0]) and name.endswith(pattern[-1])):
        return False
    for fragment in pattern:
        index = name.find(fragment)
        if index == -1:
            return False
        name = name[index + len(fragment):]
    return True


def dict_apply(dictionary: dict, function: Callable) -> dict:
    """
    Returns a copy of dict 'dictionary' where function 'function'
    was applied to all values.

    :param dictionary: dictionary to copy
    :param function: function to map
    :return: copied dictionary
    """
    return {k: function(v) for k, v in dictionary.items()}


def escape_symbols(string_to_escape: str, symbols: Union[List[str], str]) -> str:
    """
    Take a string 'string_to_escape' as input and escapes characters
    as defined in 'symbols'.

    :param string_to_escape: string where the escaping operation takes
        place
    :param symbols: list of strings to escape or string containing
        the characters to escape
    :return: escaped string
    """
    for symbol in symbols:
        string_to_escape = string_to_escape.replace(symbol, f"\\{symbol}")
    return string_to_escape


def format_str(config_path_or_dictionary: ConfigDeclarator, size: int = 200) -> str:
    """
    Format helper to shorten configs to display depending on logging level.

    :param config_path_or_dictionary: config to display
    :param size: number of characters allowed to display
    :return: the formatted string
    """
    to_return = str(config_path_or_dictionary)
    if YAECS_LOGGER.getEffectiveLevel() >= logging.INFO:
        return to_return if len(to_return) < size else f"{to_return[:size//2 - 3]} [...] {to_return[-size//2 - 3:]}"
    return to_return


def get_config_from_argv(pattern: str, fallback: Optional[ConfigInput] = None) -> List[str]:
    """
    Get paths to config files from the command line arguments.

    :param pattern: pattern to detect in sys.argv
    :param fallback: fallback value if pattern is not detected in sys.argv
    :return: the configuration
    """
    pattern_index = None
    for index, element in enumerate(sys.argv):
        if element.split("=", 1)[0] == pattern:
            pattern_index = index
    if pattern_index is not None:
        # Aggregate all CLI chunks until the next flag
        configs = []
        if "=" in sys.argv[pattern_index]:
            configs.append(sys.argv[pattern_index].split("=", 1)[1])
        for element in sys.argv[pattern_index + 1:]:
            if element.startswith("--"):
                break
            configs.append(element)
        fallback = [cfg.strip(" ") for cfg in " ".join(configs).strip().strip("[]").split(",")]
    if fallback is None:
        raise TypeError(f"The pattern '{pattern}' was not detected in sys.argv.")
    if not isinstance(fallback, list):
        fallback = [fallback]
    return [cfg for cfg in fallback if cfg]


def get_quasi_bash_sys_argv(string_to_convert: str) -> List[str]:
    """
    If a string is passed as input, process it as sys.argv would in a bash shell
    It gives exactly what sys.argv would if the script was used in a bash terminal, except that escaped '!' in quotes
    are properly escaped and the escape symbol is removed, contrary to bash (which would keep the escape for some
    obscure reason).

    :param string_to_convert: string to process
    :return: the list of strings that sys.argv would give
    """
    converted_list = [""]
    in_quotes = ""
    escaped = False
    for index, char in enumerate(string_to_convert):
        if char == "\\" and not escaped and (not in_quotes or string_to_convert[index+1] == "!"):
            escaped = True
        elif char in ['"', "'"] and not escaped:
            if not in_quotes:
                in_quotes = char
            elif in_quotes == char:
                in_quotes = ""
            else:
                converted_list[-1] += char
        elif char == " " and not in_quotes and converted_list[-1] and not escaped:
            converted_list.append("")
        elif char == "!" and not escaped:
            raise ValueError("Bash would say 'event not found', please escape the '!' character.")
        else:
            escaped = False
            converted_list[-1] += char
    if in_quotes:
        raise ValueError(f"Could not parse args : open quotations were left unclosed : {in_quotes}.")
    return converted_list


def get_order(func: Callable, default: Optional[ProcessingOrder] = Priority.INDIFFERENT) -> Optional[ProcessingOrder]:
    """
    If input function has an "order" attribute, returns it. Otherwise, returns the specified "default" value.

    :param func: function to get the order of
    :param default: default value to return if no order is found
    :return: the order value
    """
    if not hasattr(func, "yaecs_metadata") or "order" not in func.yaecs_metadata:
        return default
    return func.yaecs_metadata["order"]


def get_param_as_parsable_string(param: Any) -> str:
    """
    Gets given value as a string that can be parsed by the Configuration. The string is formatted so as to be either
    used as is in a bash shell (ie., python main.py --param_name string), or with merge_from_command_line (ie.,
    config.merge_from_command_line(f"--param_name {string}")

    :param param: parameter value to be returned as a valid string
    :raises TypeError: if the type of 'param' cannot be enforced
    :return: string usable in the command line to reproduce the value of param
    """
    container_separator = ",\\ "
    if param is None:
        return "null"
    if isinstance(param, list):
        parsable_strings = [get_param_as_parsable_string(i) for i in param]
        return f"[{container_separator.join(parsable_strings)}]"
    if isinstance(param, dict):
        parsable_strings = [f"{key}:\\ {get_param_as_parsable_string(value)}" for key, value in param.items()]
        return "{" + container_separator.join(parsable_strings) + "}"
    if isinstance(param, (int, float)) and not isinstance(param, bool):
        return format(Context(prec=20).create_decimal(repr(param)), 'f')
    if isinstance(param, str):
        string = escape_symbols(param, ['"', "'", "!", " "])
        return escape_symbols(f'"{string}"', ['"'])
    if isinstance(param, bool):
        return str(param).lower()
    raise TypeError("Provided value's type is not YAML-compatible (None, str, bool, int, float, list and dict work).")


def hook(hook_name: str) -> Callable[[Callable], Callable]:
    """
    Decorator used to keep track of registered params.

    :param hook_name: name of the hook to store
    :return: decorated function
    """
    def decorator_hook(func: Callable) -> Callable:
        if not hasattr(func, "yaecs_metadata"):
            set_function_attribute(func, "yaecs_metadata", {})
        existing_hooks = func.yaecs_metadata["hooks"] if "hooks" in func.yaecs_metadata else []
        existing_hooks += [hook_name] if hook_name not in existing_hooks else []
        func.yaecs_metadata["hooks"] = existing_hooks

        @functools.wraps(func)
        def wrapper_hook(self, *args, **kwargs):
            value = func(self, *args, **kwargs)
            self.add_currently_processed_param_as_hook(hook_name=hook_name)
            return value

        if hasattr(func, "yaecs_metadata"):
            set_function_attribute(wrapper_hook, "yaecs_metadata", func.yaecs_metadata)
        return wrapper_hook
    return decorator_hook


def is_dict_type_hint(type_hint_representer: str) -> bool:
    """
    Returns True if the type hint is a dict.

    :param type_hint_representer: type hint to check
    :return: result of the test
    """
    hint = type_hint_representer.lower().strip(" ")
    if hint == "dict":
        return True
    for fragment, pattern in TYPE_HINT_MAPPING_STARTS.items():
        if fragment.startswith("set") and hint.startswith(pattern):
            if hint.endswith(TYPE_HINT_MAPPING_ENDS[fragment]):
                return True
    return False


def is_type_valid(value: Any, config_class: type) -> bool:
    """
    Checks whether input 'value' can be saved in a YAML file by Configuration's YAML Dumper.

    :param value: value to check the type of
    :param config_class: Configuration class, which must be passed as argument to avoid circular imports :(
    :return: result of the test
    """
    if isinstance(value, list):
        return all(is_type_valid(i, config_class) for i in value)
    if isinstance(value, (Mapping, config_class)):
        return all(is_type_valid(i, config_class) for i in value.values())
    return isinstance(value, (int, float, str)) or value is None


def is_config_in_argv(pattern: str) -> bool:
    """
    Returns True if the pattern is found in sys.argv.

    :param pattern: pattern to detect in sys.argv
    :return: result of the test
    """
    try:
        _ = get_config_from_argv(pattern)
        return True
    except TypeError:
        return False


def parse_type(string_to_process: str) -> TypeHint:
    """
    Parses an input string containing the type info for a parameter into a complex type as understood by the
    Configuration.check_type function.

    :param string_to_process: string to parse for type
    :return: complex type
    """
    if not string_to_process:
        raise ValueError("Invalid type hint : empty type hint.")
    string = string_to_process.lower()
    to_return = ("root", [])
    current = []
    current_types = []
    i = 0

    def _get_sub_list(lists, path):
        list_to_get = lists
        for element in path:
            list_to_get = list_to_get[1][element]
        return list_to_get[1]

    def _increment(lists, path, value_to_add, value_type):
        list_to_incr = _get_sub_list(lists, path)
        list_to_incr.append((value_type, value_to_add))

    def _enter_list(lists, path, path_types, path_type):
        list_to_enter = _get_sub_list(lists, path)
        path.append(len(list_to_enter)-1)
        path_types.append(path_type)

    while i < len(string):
        to_find = True
        # Try to detect starts of mappings
        for type_name, fragment in TYPE_HINT_MAPPING_STARTS.items():
            if to_find and string[i:i+len(fragment)] == fragment:
                if not (fragment == "d" and string[i:i+len("dict")] == "dict"):
                    to_find = False
                    _increment(to_return, current, [], type_name)
                    _enter_list(to_return, current, current_types, type_name)
                    i += len(fragment)
        # Try to detect simple types
        for fragment, type_name in TYPE_HINT_SIMPLE_TYPES.items():
            if to_find and string[i:i+len(fragment)] == fragment:
                to_find = False
                _increment(to_return, current, type_name, "type")
                i += len(fragment)
        # Try to detect commas
        if to_find and string[i] == ",":
            to_find = False
            i += 1
        # Try to detect ends of mappings
        for type_name, fragment in TYPE_HINT_MAPPING_ENDS.items():
            if to_find and string[i:i+len(fragment)] == fragment and current_types[-1] == type_name:
                to_find = False
                current = current[:-1]
                current_types = current_types[:-1]
                i += len(fragment)
        if to_find:
            raise ValueError(f"Unexpected token at position {i} : {string_to_process}")

    if current:
        raise ValueError(f"Parsing error : unclosed brackets : {string_to_process}")

    def _struc_to_type(structured_list):
        list_to_consider = structured_list[1]
        if len(list_to_consider) != 1:
            raise ValueError("Parsing error : a source type must contain exactly 1 type (simple or complex) : "
                             f"{string_to_process}")
        if list_to_consider[0][0].startswith("type"):
            return list_to_consider[0][1]
        if list_to_consider[0][0].startswith("tuple"):
            if not list_to_consider[0][1]:
                raise ValueError(f"Parsing error : empty tuples are not allowed : {string_to_process}")
            return tuple(_struc_to_type(("", [j])) for j in list_to_consider[0][1])
        if list_to_consider[0][0].startswith("nonetuple"):
            if not list_to_consider[0][1]:
                raise ValueError(f"Parsing error : empty tuples are not allowed : {string_to_process}")
            return (None,) + tuple(_struc_to_type(("", [j])) for j in list_to_consider[0][1])
        if list_to_consider[0][0].startswith("list"):
            if not list_to_consider[0][1]:
                raise ValueError(f"Parsing error : empty lists are not allowed : {string_to_process}")
            return list(_struc_to_type(("", [j])) for j in list_to_consider[0][1])
        if list_to_consider[0][0].startswith("set"):
            return {"type": _struc_to_type(("", list_to_consider[0][1]))}
        return None

    return _struc_to_type(to_return)


def set_function_attribute(func: Callable, attribute_name: str, value: Any) -> None:
    """
    Adds an attribute to a function or method object.

    :param func: function to add the attribute to
    :param attribute_name: name of the attribute to add
    :param value: value of the attribute
    """
    try:
        setattr(func, attribute_name, value)
    except AttributeError:  # used if func is a method, to modify the underlying function
        setattr(func.__func__, attribute_name, value)


def update_state(state_descriptor: str) -> Callable[[Callable], Callable]:
    """
    Decorator used to store useful information in Configuration._state when using some recursive functions. Kind of a
    hack, but very useful to keep track of the loading state and also to debug.

    :param state_descriptor: string indicating what to store in Configuration._state
    :return: decorated function
    """

    def decorator_update_state(func: Callable) -> Callable:

        @functools.wraps(func)
        def wrapper_update_state(self, *args, **kwargs):
            # State name:
            state_to_append = state_descriptor.split(";")[0]
            for i in state_descriptor.split(";")[1:]:
                # Additional information:
                state_to_append += f";{getattr(self, i)}"
            first_arg = (args[0] if args else (kwargs[list(kwargs.keys())[0]] if kwargs else None))
            with UpdateState(state_to_append + f";arg0={first_arg}", self):
                value = func(self, *args, **kwargs)
            return value

        return wrapper_update_state

    return decorator_update_state


class UpdateState:
    """
    Context manager used to update the state of a Configuration object.
    """

    def __init__(self, state_descriptor: str, config_object: 'Configuration'):
        self._state_descriptor = state_descriptor
        self._config_object = config_object

    def __enter__(self):
        self._config_object._state.append(  # pylint: disable=protected-access
            self._state_descriptor)  # first arg of function call

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._config_object._state.pop(-1)  # pylint: disable=protected-access



"""
Reactive Reality Machine Learning Config System - ConfigSettersMixin object
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

if TYPE_CHECKING:
    from .config import Configuration

YAECS_LOGGER = logging.getLogger(__name__)


class ConfigSettersMixin:
    """ Setters Mixin class for YAECS configurations. """

    _main_config: 'Configuration'
    _pre_postprocessing_values: Dict[str, Any]
    _sub_configs_list: List['Configuration']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add_processing_function(self, param_name: str, function_to_add: Union[str, Callable], processing_type: str,
                                source: Optional[str] = None, no_duplicates: bool = True) -> None:
        """
        Adds a processing function for a param pattern.

        :param param_name: parameter(s) to which to add a postprocessing function. Expects paths with respect to the
            main config.
        :param function_to_add: postprocessing function to add, using the generic name "function" if it has no name
        :param processing_type: choose between 'pre' to add a pre-processing function or 'post' to add a post-processing
            function
        :param source: name of the source of the function
        :param no_duplicates: if True, the function will not be added if it is already in the list of processing
        """
        self.get_setter().add_processor(processor=function_to_add, pattern=param_name, processing_type=processing_type,
                                        source=source, no_duplicates=no_duplicates, container=self)

    def remove_value_before_postprocessing(self, name: str) -> None:
        """
        Function used for bookkeeping : it remove a parameter from the pre-post-processing archive.

        :param name: name of the parameter using the dot convention
        """
        if name in self._pre_postprocessing_values:
            del self._pre_postprocessing_values[name]

    def save_value_before_postprocessing(self, name: str, value: Any) -> None:
        """
        Function used for bookkeeping : it saves the value a parameter had before its post-processing.

        :param name: name of the parameter using the dot convention
        :param value: value of the parameter before post-processing
        """
        if name not in self._pre_postprocessing_values:
            self._pre_postprocessing_values[name] = value

    def set_variation_name(self, value: Optional[str]) -> None:
        """
        Sets the variation name across the entire config object. Calling this for a sub-config will also affect the main
        config and all other sub-configs.

        :param value: value of the new variation name
        """
        object.__setattr__(self._main_config, "_variation_name", value)
        for subconfig in self._main_config.get_sub_configs(deep=True):
            object.__setattr__(subconfig, "_variation_name", value)



""" YAMLScanner class to pre-parse YAML files before feeding it to the config """

import logging
import re
from typing import Any, Dict, Optional, Type

import yaml

from ..yaecs_utils import parse_type, YAML_EXPRESSIONS

YAECS_LOGGER = logging.getLogger(__name__)


class YAMLScanner:
    """ YAMLScanner class to pre-parse YAML files before feeding it to the config """

    def __init__(self, yaml_path: str):
        self.path: str = yaml_path
        self.params: Dict[str, Any] = {}
        self.type_hints: Dict[str, str] = {}
        self.processing_functions: Dict[str, str] = {}
        self.state: Dict[str, Any] = {
            "currently_processed_path": [],
            "last_non_scalar": 0,
            "last_ended_mapping": -1,
            "sequence_depth": [],
        }
        with open(yaml_path, encoding='utf-8') as yaml_file:
            list(yaml.load_all(yaml_file, Loader=self._get_yaml_loader()))

    def resolve_node(self, yaml_loader, tag, node):
        """ Resolve a YAML node to a Python object. """

        with NoTag(yaml_loader):
            if isinstance(node, yaml.ScalarNode):
                if node.value == "":
                    def _can_be_str(parsed_type):
                        if parsed_type is str:
                            return True
                        if isinstance(parsed_type, tuple):
                            if str in parsed_type:
                                return True
                            return any(_can_be_str(t) for t in parsed_type)
                        return False
                    if tag == "!:no-tag:" or _can_be_str(parse_type(tag[len("!type:"):])):
                        return yaml_loader.default_yaml_constructors["tag:yaml.org,2002:str"](yaml_loader, node)
                for key, value in YAML_EXPRESSIONS.items():
                    if value.match(node.value):
                        return yaml_loader.default_yaml_constructors[f"tag:yaml.org,2002:{key}"](yaml_loader, node)
                return yaml_loader.construct_scalar(node)
            if isinstance(node, yaml.SequenceNode):
                return yaml_loader.construct_sequence(node, deep=True)
            if isinstance(node, yaml.MappingNode):
                return yaml_loader.construct_mapping(node, deep=True)
            raise ValueError(f"Unsupported node type {type(node)}.")

    def update_state(self, yaml_loader, tag, node):
        """ Updates the state of the scanner as it parses the YAML file. """

        self.state["is_param_tag"] = bool(yaml_loader.constructed_objects) or tag.lower() == "!:no-tag:"
        self.state["resolving_recursive_param"] = yaml_loader.DEFAULT_MAPPING_TAG == "!:no-tag:"
        complex_added = 0
        for value in list(yaml_loader.constructed_objects.keys())[::-1]:
            if isinstance(value, (yaml.MappingNode, yaml.SequenceNode)):
                complex_added += 1
            else:
                break
        node_type = "scalar" if isinstance(node, yaml.ScalarNode) else "sequence" if isinstance(
            node, yaml.SequenceNode) else "mapping"

        if node_type == "sequence":
            self.state["sequence_depth"].append("sequence")
        if self.state["sequence_depth"]:
            if node_type == "mapping":
                self.state["sequence_depth"].append("mapping")
            if complex_added and len(yaml_loader.constructed_objects) != self.state["last_ended_mapping"]:
                self.state["sequence_depth"] = self.state["sequence_depth"][:-complex_added]
                self.state["last_ended_mapping"] = len(yaml_loader.constructed_objects)
        if node_type == "mapping" or complex_added:
            self.state["last_non_scalar"] = len(yaml_loader.constructed_objects)

        key_counter = len(yaml_loader.constructed_objects) - self.state["last_non_scalar"]
        if self.state["sequence_depth"] and self.state["sequence_depth"][-1] == "sequence":
            self.state["is_key_node"] = False
        else:
            self.state["is_key_node"] = node_type == "scalar" and self.state["is_param_tag"] and key_counter % 2 == 0

        if not self.state["is_key_node"] and not self.state["is_param_tag"]:
            self.state["last_ended_mapping"] = -1

    def _get_yaml_loader(self) -> Type[yaml.FullLoader]:
        """ Used to get a custom YAML loader capable of parsing config tags. """

        def generic_constructor(yaml_loader, tag, node):

            self.update_state(yaml_loader, tag, node)

            if self.state["is_key_node"]:
                return self.resolve_node(yaml_loader, tag, node)

            if self.state["is_param_tag"]:
                if self.state["resolving_recursive_param"]:
                    return self.resolve_node(yaml_loader, tag, node)

                name = yaml_loader.constructed_objects[list(yaml_loader.constructed_objects.keys())[-1]]
                check_valid_param_name(name, f"Invalid name '{name}' found in file {self.path} : " + "{issue}.")
                full_name = ".".join(self.state["currently_processed_path"] + [name])

                if tag.lower() != "!type:config":
                    if tag.lower().startswith("!type:"):
                        self.type_hints[full_name] = tag[len("!type:"):]
                    elif tag != "!:no-tag:" and not tag.startswith("tag:yaml.org,2002:"):
                        self.processing_functions[full_name] = tag[1:].split(",")

                    self.params[full_name] = self.resolve_node(yaml_loader, tag, node)
                    return self.params[full_name]

            else:
                if tag.lower() == "!type:config":
                    # If root of the YAML file but no provided name > assume dict
                    return yaml_loader.construct_mapping(node, deep=True)
                # If root of the YAML file and a name is provided > use first part as config name
                name = tag[1:]
                check_valid_param_name(name, f"Invalid name '{name}' found in file {self.path} : " + "{issue}.")

            with InSubConfig(self, name):
                value = yaml_loader.construct_mapping(node)

            return value

        loader = yaml.FullLoader
        if not hasattr(loader, "default_yaml_constructors"):
            loader.default_yaml_constructors = dict(loader.yaml_constructors)
            loader.yaml_constructors = {}
        loader.DEFAULT_MAPPING_TAG = "!type:config"
        loader.DEFAULT_SCALAR_TAG = "!:no-tag:"
        loader.DEFAULT_SEQUENCE_TAG = "!:no-tag:"
        yaml.add_multi_constructor("", generic_constructor, Loader=loader)

        return loader


def check_valid_param_name(name: str, message: Optional[str] = None) -> None:
    """ Raise relevant errors in the input name is not a valid sub-config name. """
    issue = None
    if not name:
        issue = "sub-config names cannot be empty"
    if not re.match(r'^[A-Za-z_*]', name):
        issue = "sub-config names must start with a letter or an underscore"
    if not re.match(r'^[A-Za-z0-9._*]*$', name):
        issue = "sub-config names can only contain letters, numbers, dots and underscores"
    if name.endswith('.'):
        issue = "sub-config names cannot end with a dot"
    if issue:
        raise ValueError(message.format_map({"issue": issue}))


class NoTag:
    """ Context manager to temporarily set the default mapping of a loader to "!:no-tag:" """

    def __init__(self, loader):
        self.loader = loader
        self.saved_default_mapping_tag = None

    def __enter__(self):
        self.saved_default_mapping_tag = self.loader.DEFAULT_MAPPING_TAG
        self.loader.DEFAULT_MAPPING_TAG = "!:no-tag:"

    def __exit__(self, *args):
        self.loader.DEFAULT_MAPPING_TAG = self.saved_default_mapping_tag


class InSubConfig:
    """ Context manager to temporarily append a name to the currently processed path. """

    def __init__(self, scanner, name):
        self.scanner = scanner
        self.name = name

    def __enter__(self):
        self.scanner.state["currently_processed_path"].append(self.name)

    def __exit__(self, *args):
        self.scanner.state["currently_processed_path"].pop(-1)



""" This file implements a Setter class that is used to process and set parameters. It handles checking types based on
type hints, pre-processing, and post-processing. It can store functions and strings referring to methods, in which case
a method container should be passed to resolve those methods. """
from collections.abc import Iterable
import logging
from numbers import Real
from typing import Any, Callable, Dict, List, Optional, Union

from ..yaecs_utils import (Priority, ProcessingFunction, ProcessingFunctions, ProcessingOrder, TypeHint, UpdateState,
                           check_type, compare_string_pattern, is_type_valid, parse_type)

YAECS_LOGGER = logging.getLogger("yaecs")


class Setter:
    """ Processes parameters to set them in a container. Handles type checking based on type hints, pre-processing, and
    post-processing. Can store functions and strings referring to methods, in which case a method container should be
    passed to resolve those methods. """

    def __init__(self, registered_methods: dict, do_not_pre_process: bool = False, do_not_post_process: bool = False,
                 default_order: ProcessingOrder = Priority.INDIFFERENT, verbose: bool = True):
        """
        Initializes the ParamProcessor object.

        :param registered_methods: dictionary of registered methods and their metadata
        :param do_not_pre_process: whether to skip pre-processing
        :param do_not_post_process: whether to skip post-processing
        :param default_order: default order to use if no order is specified for a processor
        :param verbose: whether to print verbose messages
        """
        self.registered_methods: dict = registered_methods
        self.default_order: ProcessingOrder = default_order
        self.verbose: bool = verbose
        self.processes: List[str] = ([] if do_not_pre_process else ["pre"]) + ([] if do_not_post_process else ["post"])
        self.processors: Dict[str, List['Processor']] = {
            "pre": [],
            "post": [],
        }

    def __call__(self, names: Dict[str, str], values: Dict[str, Any], processing_type: str, container: object,
                 only_set_processed_parameters: bool = False) -> None:
        """
        Processes a parameter based on the processing type.

        :param parameters: parameters to process with names as keys
        :param processing_type: type of processing to perform
        :param container: object containing methods to resolve processors passed as strings, and where to set the values
        :param only_set_processed_parameters: whether to only set the parameters for which at least one processor
        applied
        :return: processed value
        """
        if processing_type not in self.processors:
            raise ValueError(f"Unknown processing_type : '{processing_type}'. "
                             f"Valid types are {list(self.processors.keys())}.")

        processed = []
        if processing_type in self.processes:
            processors = [processor for processor in self.processors[processing_type] if processor.applies(values)]
            processors.sort(key=lambda processor: processor.order)
            processors = self._resolve_type_hints(processors)

            processed_values = dict(values)
            for processor in processors:
                for name, value in processed_values.items():
                    if processor.applies(name):
                        with UpdateState(f"processing;{container.get_name()};arg0={name}", container):
                            processed_values[name] = processor(name, value, container=container)
                        self._set_value(names[name], processed_values[name], container)
                        if name not in processed:
                            processed.append(name)

        if not only_set_processed_parameters:
            for name, value in values.items():
                if name not in processed:
                    self._set_value(names[name], value, container)

        if processing_type == "post" and processed and self.verbose:
            YAECS_LOGGER.info(f"Performed post-processing for modified parameters {processed}.")
        if processing_type == "pre":
            for name, value in values.items():
                if not is_type_valid(values[name], container.__class__):
                    raise RuntimeError(f"ERROR while pre-processing param '{name}' : pre-processing functions that "
                                       "change the type of a param to a non-native YAML type are forbidden because "
                                       "they cannot be saved. Please use a parameter post-processing instead.")

    def add_type_hint(self, type_hint: str, pattern: str, source: Optional[str] = None,
                      no_duplicates: bool = False) -> None:
        """
        Adds a type hint.

        :param type_hint: type hint to add
        :param pattern: pattern that the parameter name must match to be checked
        :param source: if provided, information about the source that added the processor
        :param no_duplicates: whether to check for duplicates before adding
        """
        new_processor = Processor(
            processor=type_hint,
            pattern=pattern,
            is_type_check=True,
            source=source,
        )
        if not no_duplicates or new_processor not in self.processors["pre"]:
            self.processors["pre"].append(new_processor)

    def add_processor(self, processor: ProcessingFunctions, pattern: str, processing_type: Optional[str] = None,
                      order: Optional[ProcessingOrder] = None, source: Optional[str] = None,
                      no_duplicates: bool = False, container: Optional[object] = None) -> None:
        """
        Adds a processor.

        :param processor: processing function
        :param pattern: pattern that the parameter name must match to be processed
        :param processing_type: type of processing to add the processor to
        :param order: order of the processing function
        :param source: if provided, information about the source that added the processor
        :param no_duplicates: whether to check for duplicates before adding
        :param container: if passed, callables passed are stored as strings if they are methods of the container
        """
        # Recover the processor's metadata
        metadata = None
        if isinstance(processor, str):
            if processor in self.registered_methods:
                metadata = self.registered_methods[processor]
            elif container is not None and processor in dir(container):
                candidate = getattr(container, processor)
                if hasattr(candidate, "yaecs_metadata"):
                    metadata = candidate.yaecs_metadata
            else:
                for data in self.registered_methods.values():
                    if "name" in data and data["name"] == processor:
                        metadata = data
                        break
        else:
            metadata = getattr(processor, "yaecs_metadata", None)

        # If the processing type is not provided, infer it from the metadata
        if processing_type is None and (metadata is None or "processing_type" not in metadata):
            source_message = "" if source is None else f" (added from {source})"
            raise ValueError("Processing type was not provided and could not be inferred from metadata for processor "
                             f"'{processor}'{source_message} defined for pattern '{pattern}'.")
        if processing_type is None:
            processing_type = metadata["processing_type"]

        # If the processor is a method of the container, store it as a string
        if container is not None and isinstance(processor, Callable):
            if hasattr(processor, "yaecs_metadata"):
                name = processor.yaecs_metadata.get("name", getattr(processor, "__name__", "unknown_function"))
            else:
                name = getattr(processor, "__name__", "unknown_function")
            if name in dir(container):
                candidate = getattr(container, name)
                processor_metadata = getattr(processor, "yaecs_metadata", {})
                candidate_metadata = getattr(candidate, "yaecs_metadata", {})
                if isinstance(candidate, Callable) and processor_metadata == candidate_metadata:
                    processor = name
                    processor_metadata["name"] = name
                    self.registered_methods[name] = processor_metadata
                    metadata = processor_metadata

        # Create the new processor
        new_processor = Processor(
            processor=processor,
            pattern=pattern,
            order=order,
            default_order=self.default_order,
            is_type_check=False,
            source=source,
            metadata=metadata,
        )

        # Add the new processor
        if not no_duplicates or new_processor not in self.processors[processing_type]:
            self.processors[processing_type].append(new_processor)
            if metadata and "input_type" in metadata:
                self.add_type_hint(metadata["input_type"], pattern, source=f"method[{processor}]")
            if new_processor.metadata is not None and "processing_type" in new_processor.metadata:
                advised_type = new_processor.metadata["processing_type"]
                if advised_type != processing_type:
                    YAECS_LOGGER.warning(f"WARNING : processor '{new_processor.name}' is recommended to use as "
                                         f"{advised_type}-processing function, but was declared as "
                                         f"{processing_type}-processing function.")

    def bulk_add_type_hints(self, type_hints: Dict[str, str], source: Optional[str] = None,
                            no_duplicates: bool = False) -> None:
        """
        Adds multiple type hints at once.

        :param type_hints: dictionary of type hints to add
        :param source: if provided, information about the source that added the type hints
        :param no_duplicates: whether to check for duplicates before adding
        """
        for pattern, type_hint in type_hints.items():
            self.add_type_hint(type_hint=type_hint, pattern=pattern, source=source, no_duplicates=no_duplicates)

    def bulk_add_processors(self, processors: Dict[str, ProcessingFunctions], processing_type: Optional[str] = None,
                            source: Optional[str] = None, no_duplicates: bool = False,
                            container: Optional[object] = None) -> None:
        """
        Adds multiple processors at once.

        :param processors: dictionary of processors to add
        :param processing_type: type of processing to add the processors to. If None, must be inferrable from metadata
        :param source: if provided, information about the source that added the processors
        :param no_duplicates: whether to check for duplicates before adding
        :param container: if passed, callables passed are stored as strings if they are methods of the container
        """
        for pattern, processor in processors.items():
            if not isinstance(processor, (Callable, Iterable)):
                source_message = "" if source is None else f" (added from {source})"
                type_message = "" if processing_type is None else f"{processing_type}-"
                raise TypeError(f"Invalid {type_message}processing functions{source_message} defined for pattern "
                                f"'{pattern}' : the function should be declared as either a function or an iterable of "
                                "functions, optionally containing one order value.")

            order = None
            if isinstance(processor, Iterable) and not isinstance(processor, str):
                if any(not isinstance(element, (Callable, str, Real, Priority)) for element in processor):
                    source_message = "" if source is None else f" (added from {source})"
                    type_message = "" if processing_type is None else f"{processing_type}-"
                    raise TypeError(f"Invalid {type_message}processing functions{source_message} defined for "
                                    f"pattern '{pattern}' : if function is declared as iterable, only functions and "
                                    "one order value can be provided.")
                order = [o for o in processor if isinstance(o, (Real, Priority))]
                if len(order) > 1:
                    source_message = "" if source is None else f" (added from {source})"
                    type_message = "" if processing_type is None else f"{processing_type}-"
                    raise ValueError(f"Ambiguous order for {type_message}processing functions{source_message} "
                                     f"defined for pattern '{pattern}' : multiple orders defined ({order}).")
                processor = [p for p in processor if isinstance(p, (Callable, str))]
                order = order[0] if order else None

            else:
                processor = [processor]

            for p in processor:
                self.add_processor(processor=p, pattern=pattern, processing_type=processing_type, order=order,
                                   source=source, no_duplicates=no_duplicates, container=container)

    def set_post_processing(self, value: bool = True):
        """
        Sets whether to post-process.

        :param value: whether to post-process
        """
        self._set_processing(value, "post")

    def set_pre_processing(self, value: bool = True):
        """
        Sets whether to pre-process.

        :param value: whether to pre-process
        """
        self._set_processing(value, "pre")

    def _set_processing(self, value: bool, processing_type: str):
        """
        Sets whether to process for a given processing type.

        :param value: whether to process
        :param processing_type: type of processing to set
        """
        if value and processing_type not in self.processes:
            self.processes.append(processing_type)
        elif not value and processing_type in self.processes:
            self.processes.remove(processing_type)

    def _set_value(self, name, value, container) -> None:
        """
        Sets the value of a parameter.

        :param name: name of the parameter
        :param value: value to set
        :param container: object containing the parameter
        """
        sub_container = ".".join(name.split(".")[:-1])
        param_name = name.split(".")[-1]
        object.__setattr__((container.get_main_config()[sub_container] if sub_container else container),
                           param_name, value)

    def _resolve_type_hints(self, processors: List['Processor']) -> List['Processor']:
        """
        Remove type hints from the processors if they are erroneous.

        :param processors: processors to resolve
        :return: resolved processors
        """
        type_hints = [processor for processor in processors if processor.is_type_check]
        if not type_hints:
            return processors

        valid_auto_type_hint = []
        for type_hint in type_hints:
            source = "" if type_hint.source is None else type_hint.source
            is_auto_type_hint = source.startswith("method[") and source.endswith("]")
            if not is_auto_type_hint:
                return [type_hint] + processors[len(type_hints):]
            if source[len("method["):len("]")] == processors[0].processor:
                valid_auto_type_hint = [type_hint]

        return valid_auto_type_hint + processors[len(type_hints):]


class Processor:
    """ Class used to store a processing function. """
    MIN_ORDER = -1000000

    def __init__(self, processor: ProcessingFunction, pattern: str, order: Optional[ProcessingOrder] = None,
                 default_order: ProcessingOrder = Priority.INDIFFERENT, is_type_check: bool = False,
                 source: Optional[str] = None, metadata: Optional[dict] = None):
        """
        Initializes the Processor object.

        :param processor: processing function
        :param pattern: pattern that the parameter name must match to be processed
        :param order: order of the processing function
        :param default_order: order to use if no order is specified
        :param is_type_check: whether the function is a type to be checked
        :param source: if type check was added automatically, which method it was added from
        :param metadata: dictionary of metadata about the processor
        """
        self.pattern: str = pattern
        self.is_type_check: bool = is_type_check
        self.source: Optional[str] = source
        self.metadata: dict = {} if metadata is None else metadata
        if not isinstance(processor, str) and hasattr(processor, "yaecs_metadata"):
            self.metadata.update(processor.yaecs_metadata)
        self.name = self._resolve_name(processor)
        self.processor: Union[ProcessingFunction, TypeHint] = parse_type(processor) if self.is_type_check else processor
        self.order: ProcessingOrder = self._resolve_order(order, default_order)

    def __call__(self, name: str, old_value: Any, container: Optional[object] = None) -> Any:
        """
        Calls the processing function.

        :param name: name of the parameter
        :param old_value: argument to pass to the function
        :param container: object containing methods to resolve processors passed as strings
        :return: return value of the function
        """
        if not self.applies(name):
            return old_value
        function = self._resolve(container, name)
        try:
            return function(old_value)
        except Exception:
            YAECS_LOGGER.error(f"ERROR while processing param '{name}'.")
            raise

    def __eq__(self, other: 'Processor') -> bool:
        """
        Checks whether two processors are equal.

        :param other: processor to compare
        :return: whether the processors are equal
        """
        return (self.pattern == other.pattern and self.name == other.name and self.order == other.order
                and self.source == other.source)

    def applies(self, params_or_param_name: Union[Dict[str, Any], str]) -> bool:
        """
        Checks whether the processor applies to the parameter(s).

        :param params_or_param_name: parameter(s) to check
        :return: whether the processor applies
        """
        if isinstance(params_or_param_name, str):
            return compare_string_pattern(params_or_param_name, self.pattern)
        return any(compare_string_pattern(name, self.pattern) for name in params_or_param_name.keys())

    def _resolve(self, container: object, name: str) -> Callable:
        """
        If the processor is a string, this method resolves the string to a method of the container.

        :param container: object containing methods to resolve processors passed as strings
        :return: resolved function
        """
        if self.is_type_check:
            return check_type(self.processor, name)
        if isinstance(self.processor, str):
            if (self.name not in dir(container)
                    or not isinstance(getattr(container, self.name), Callable)):
                source_message = "" if self.source is None else f" (added from {self.source})"
                raise ValueError(f"Method '{self.name}'{source_message} not found in {container}.")
            return getattr(container, self.name)
        return self.processor

    def _resolve_name(self, processor: ProcessingFunction) -> str:
        """
        Resolves the name of the processor.

        :param processor: processor to resolve the name of
        :return: resolved name
        """
        if "name" in self.metadata:
            return self.metadata["name"]
        return processor if isinstance(processor, str) else getattr(processor, "__name__", "unknown_function")

    def _resolve_order(self, order_overwrite: Optional[ProcessingOrder],
                       default_order: ProcessingOrder = Priority.INDIFFERENT) -> ProcessingOrder:
        """
        Resolves the order of the function.

        :param order_overwrite: order to use over all other orders
        :param default_order: default order to use if no order is specified
        :return: resolved order
        """
        if self.is_type_check:
            return self.MIN_ORDER - 1

        order = order_overwrite if order_overwrite is not None else self.metadata.get("order", default_order)
        return max(order, self.MIN_ORDER)



"""
Reactive Reality Machine Learning Config System - _ConfigurationBase object
Copyright (C) 2022  Reactive Reality

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import copy
import logging
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

import yaml

from ..yaecs_utils import (ConfigDeclarator, NoValue,
                           compare_string_pattern, format_str, get_quasi_bash_sys_argv, is_dict_type_hint, update_state)
from .config_convenience import ConfigConvenienceMixin
from .config_getters import ConfigGettersMixin
from .config_hooks import ConfigHooksMixin
from .config_processing_functions import ConfigProcessingFunctionsMixin
from .config_setters import ConfigSettersMixin
from .setter import Setter
from .yaml_scanner import YAMLScanner

if TYPE_CHECKING:
    from .config import Configuration

YAECS_LOGGER = logging.getLogger(__name__)


class _ConfigurationBase(ConfigHooksMixin, ConfigGettersMixin, ConfigSettersMixin, ConfigConvenienceMixin,
                         ConfigProcessingFunctionsMixin):
    """ Base class for YAECS configurations. Defines its basic behaviour, such as creation and merging operations,
    including processing- and type checking-related logic, but not its constructors (see Configuration class for those,
    and its docstring for more details about the composition of the Configuration superclass). """

    config_metadata: dict
    parameters_pre_processing: Callable[[], Dict[str, Callable]]
    parameters_post_processing: Callable[[], Dict[str, Callable]]
    _get_instance: Callable
    _get_tagged_methods_info: Callable[[], List[Tuple[Union[str, Callable]]]]
    _main_config: 'Configuration'
    _methods: List[str]
    _nesting_hierarchy: List[str]
    _operating_creation_or_merging: bool
    _protected_attributes: List[str]
    _state: List[str]
    _verbose: bool

    def __init__(self, from_argv: str = "", do_not_pre_process: bool = False, do_not_post_process: bool = False):
        """
        Should never be called directly by the user. Please use one of the constructors defined for the Configuration
        class instead, or the utils.make_config convenience function.

        :param from_argv: pattern used to find the config in the command line arguments, or "" if not applicable
        :param do_not_pre_process: if true, pre-processing is deactivated in this initialization
        :param do_not_post_process: if true, post-processing is deactivated in this initialization
        :raises ValueError: if the overwriting regime is not valid
        :return: none
        """

        # PROTECTED ATTRIBUTES
        if self._is_main_config():
            self._modified_buffer = []
            self._setter = Setter(registered_methods=self._get_tagged_methods_info(),
                                  do_not_post_process=do_not_post_process, do_not_pre_process=do_not_pre_process,
                                  verbose=self._verbose)
            for process_type in ["pre", "post"]:
                source = f"parameters_{process_type}_processing"
                self._setter.bulk_add_processors(processors=getattr(self, source)(),
                                                 processing_type=process_type, source=source, container=self)
        self._former_saving_time = None
        self._from_argv = from_argv
        self._pre_postprocessing_values = {}
        self._reference_folder = None
        self._sub_configs_list = []
        self._was_last_saved_as = None
        super().__init__()

    def __getitem__(self, item) -> Any:
        if "." in item and "*" not in item:
            sub_config_name = ("___"
                               + item.split(".")[0] if item.split(".")[0] in self._methods else item.split(".")[0])
            sub_config = getattr(self, sub_config_name)
            if not isinstance(sub_config, _ConfigurationBase):
                did_you_mean_message = self._did_you_mean(sub_config_name, filter_type=self.__class__)
                raise TypeError(f"As the parameter '{sub_config_name}' is not a sub-config"
                                f", it cannot be accessed.\n{did_you_mean_message}")
            return sub_config[item.split(".", 1)[1]]
        return getattr(self, "___" + item if item in self._methods else item)

    def __setattr__(self, key, value) -> None:
        if (self.is_in_operation() or self._main_config.is_in_operation()
                or self.config_metadata["overwriting_regime"] == "unsafe"):
            object.__setattr__(self, key, value)
        elif self.config_metadata["overwriting_regime"] == "auto-save":
            self._manual_merge({key: value}, source='code')
        elif self.config_metadata["overwriting_regime"] == "locked":
            raise RuntimeError("Overwriting params in locked configs is not allowed.")
        else:
            raise ValueError(f"No behaviour determined for value '{self.config_metadata['overwriting_regime']}' of "
                             "parameter 'overwriting_regime'.")

    def __getattribute__(self, item) -> Any:
        try:
            return object.__getattribute__(self, item)
        except AttributeError as exception:
            if not item.startswith("_") and not any(state.startswith("setup") for state in self._state):
                raise AttributeError(f"Unknown parameter of the configuration : '{item}'.\n"
                                     f"{self._did_you_mean(item)}") from exception
            raise AttributeError from exception

    def __iter__(self):
        return iter(self._get_user_defined_attributes())

    @update_state("init_from_config;_name")
    def init_from_config(self, config_path_or_dict: ConfigDeclarator) -> None:
        """
        Entrypoint for all methods trying to get any value from outside the config to inside the config. This
        includes creating new parameters when creating the config or merging existing parameters after the creation.
        Users should only use this to merge or create parameters during a creation or merge operation (for instance in a
        processing function). If you want to merge parameters in your main code, outside a constructor or other
        operation, please use self.merge.

        :param config_path_or_dict: path or dictionary for the config to merge
        """
        if config_path_or_dict is None:
            config_path_or_dict = {}
        if isinstance(config_path_or_dict, str):
            config_path_or_dict = self._scan_yaml_files(config_path_or_dict)

        for item in config_path_or_dict.items():
            self._process_item_to_merge_or_add(item)

    def merge(self, config_path_or_dictionary: ConfigDeclarator, do_not_pre_process: bool = False,
              do_not_post_process: bool = False) -> None:
        """
        Merges provided config path of dictionary into the current config.

        :param config_path_or_dictionary: path or dictionary for the config to merge
        :param do_not_pre_process: if true, pre-processing is deactivated in this initialization
        :param do_not_post_process: if true, post-processing is deactivated in this initialization
        """
        self._manual_merge(config_path_or_dictionary=config_path_or_dictionary, do_not_pre_process=do_not_pre_process,
                           do_not_post_process=do_not_post_process)

    def merge_from_command_line(self, to_merge: Optional[Union[List[str], str]] = None,
                                do_not_pre_process: bool = False, do_not_post_process: bool = False) -> None:
        """
        Formerly used to manually merge the command line arguments into the config, which is now done automatically and
        thus should no longer be done manually. Can still be used to manually merge a string emulating command line
        arguments.

        :param to_merge: if specified, merges this string or list of strings instead of the sys.argv list of strings
        :param do_not_pre_process: if true, pre-processing is deactivated in this initialization
        :param do_not_post_process: if true, post-processing is deactivated in this initialization
        """
        if self._verbose and to_merge is None:
            # TODO handle all warnings with a function _warn that logs and stores messages for future prints
            YAECS_LOGGER.warning("WARNING : merge_from_command_line is now deprecated and will automatically start "
                                 "after using any constructor.\nYou can remove the 'config.merge_from_command_line()' "
                                 "line from your code now :) it's redundant.")
        to_merge = self._gather_command_line_dict(to_merge)
        if to_merge:
            self._manual_merge(to_merge, do_not_pre_process=do_not_pre_process, do_not_post_process=do_not_post_process,
                               source='command line')

    def _find_path(self, path: str) -> str:
        """ Used to find a config from its (potentially relative) path, because it might be ambiguous relative to where
        it should be looked for. Probably very improvable. """
        def _get_path(path_to_check):
            if os.path.exists(path_to_check):
                return os.path.abspath(path_to_check)
            if path_to_check.endswith(".yaml") or path_to_check.endswith(".yml"):
                to_check = ".".join(path_to_check.split(".")[:-1])
            else:
                to_check = path_to_check
            there = [os.path.exists(to_check + ".yaml"), os.path.exists(to_check + ".yml"), os.path.exists(to_check)]
            if sum(there) == 0:
                return None
            if sum(there) == 1:
                return os.path.abspath([to_check + ".yaml", to_check + ".yml", to_check][there.index(True)])
            there.pop(there.index(False))
            raise RuntimeError(f"Ambiguity for provided path '{path_to_check}' : detected two possible paths :\n"
                               f"   - {there[0]}\n   - {there[1]}")

        # If the path is absolute, use it...
        if os.path.isabs(path):
            processed_path = _get_path(path)
            if processed_path is not None:
                self._reference_folder = str(Path(processed_path).parents[0])
                return processed_path

        # ... if not, search relatively to some reference folders.
        else:
            possibilities = []
            last_is_current = False

            # First check relatively to parent configs' directories...
            for config in reversed(self.config_metadata["config_hierarchy"]):
                if isinstance(config, str):
                    relative_path = os.path.join(Path(config).parents[0], path)
                    absolute_path = _get_path(relative_path)
                    if absolute_path is not None and absolute_path not in possibilities:
                        possibilities.append(absolute_path)

            # ... then also check the current reference folder since
            # the config hierarchy is not always up-to-date...
            if self._reference_folder is not None:
                relative_path = os.path.join(self._reference_folder, path)
                absolute_path = _get_path(relative_path)
                if absolute_path is not None and absolute_path not in possibilities:
                    possibilities.append(absolute_path)
            if self._main_config.get_reference_folder() is not None:
                relative_path = os.path.join(self._main_config.get_reference_folder(), path)
                absolute_path = _get_path(relative_path)
                if absolute_path is not None and absolute_path not in possibilities:
                    possibilities.append(absolute_path)

            # ... and finally, check relatively to the current
            # working directory.
            absolute_path = _get_path(path)
            if absolute_path is not None and absolute_path not in possibilities:
                last_is_current = True
                possibilities.append(absolute_path)

            if len(possibilities) > 1:
                if path.endswith(".yaml"):
                    filtered = [p for p in possibilities if p.endswith(".yaml")]
                elif path.endswith(".yml"):
                    filtered = [p for p in possibilities if p.endswith(".yml")]
                else:
                    filtered = [p for p in possibilities if not p.endswith(".yaml") and not p.endswith(".yml")]
            else:
                filtered = possibilities

            if len(filtered) > 1:
                YAECS_LOGGER.warning(f"WARNING : Multiple matches for path {path}. '{filtered[0]}' will be used.\n"
                                     f"All matches : {filtered}.")
            if filtered:
                if last_is_current and filtered[0] == possibilities[-1]:
                    self._reference_folder = str(Path(absolute_path).parents[0])
                return filtered[0]

        raise FileNotFoundError(f"ERROR : no YAML file found at path '{path}'.")

    def _manual_merge(self, config_path_or_dictionary: ConfigDeclarator, do_not_pre_process: bool = False,
                      do_not_post_process: bool = False, source: str = 'config',
                      ) -> None:
        """ This method is called whenever a merge is done by the user, and not by the config creation process. It
        simply calls _merge with some additional bookkeeping. """
        self._merge(config_path_or_dictionary=config_path_or_dictionary, do_not_pre_process=do_not_pre_process,
                    do_not_post_process=do_not_post_process, source=source)
        self._post_process_modified_parameters()
        self.get_setter().set_post_processing(True)
        if self._main_config.config_metadata["overwriting_regime"] == "auto-save":
            if self._main_config.get_save_file() is not None:
                self._main_config.save()

    @update_state("merging;_name")
    def _merge(self, config_path_or_dictionary: ConfigDeclarator, do_not_pre_process: bool = False,
               do_not_post_process: bool = False, source: str = 'config') -> None:
        """ Method handling all merging operations to call init_from_config with the proper bookkeeping. """
        if self._is_main_config():
            object.__setattr__(self, "_operating_creation_or_merging", True)
            if self._verbose:
                YAECS_LOGGER.info(f"Merging from {source} : {format_str(config_path_or_dictionary)}")
            self.get_setter().set_post_processing(not do_not_post_process)
            self.get_setter().set_pre_processing(not do_not_pre_process)
            self.init_from_config(config_path_or_dictionary)
            self.config_metadata["config_hierarchy"].append(config_path_or_dictionary)
            self.get_setter().set_pre_processing(True)
            self._operating_creation_or_merging = False
        else:
            if isinstance(config_path_or_dictionary, str):
                config_path_or_dictionary = self._scan_yaml_files(config_path_or_dictionary)

            if config_path_or_dictionary is not None:
                config_path_or_dictionary = {self._get_full_path(a): b for a, b in config_path_or_dictionary.items()}

            self._main_config._merge(  # pylint: disable=protected-access
                config_path_or_dictionary=config_path_or_dictionary,
                do_not_pre_process=do_not_pre_process, do_not_post_process=do_not_post_process, source=source
            )

    @update_state("working_on;_name")
    def _process_item_to_merge_or_add(self, item: Tuple[str, Any]) -> None:
        """ Method called by init_from_config to merge or add a given key, value pair. """
        key, value = item

        # Process metadata. If there is metadata, treat the rest of the merge as "loading a saved file"... (which will
        # deactivate the parameter pre-processing for this merge).
        if key == "config_metadata":
            former_saving_time, regime, variation = self._parse_metadata(value)
            self._former_saving_time = former_saving_time
            self.config_metadata["overwriting_regime"] = regime
            self.set_variation_name(variation)
            self.get_setter().set_pre_processing(False)
            return

        # ...do not accept other protected attributes to be merged...
        if key in self._protected_attributes:
            raise RuntimeError(f"Error : '{key}' is a protected name and cannot be used as a parameter name.")

        # ... otherwise, process the data normally :

        # If we are merging a parameter into a previously defined config...
        if not any(state.startswith("setup") for state in self._state):
            self._merge_item(key, value)

        # ... or if we are creating a config for the first time and are adding non-existing parameters to it
        else:
            self._add_item(key, value)

    def _merge_item(self, key: str, value: Any) -> None:
        """ Method called by _process_item_to_merge_or_add if the value should be merged and not added (ie., any time
        after the default config has been set up). This method ultimately performs all merges in the config. """

        if "*" in key:
            to_merge = {}
            for param in self.get_parameter_names(deep=True):
                if compare_string_pattern(param, key):
                    to_merge[param] = value
            if self._verbose:
                if not to_merge:
                    YAECS_LOGGER.warning(f"WARNING : parameter '{key}' will be ignored : it does not match any existing"
                                         " parameter.")
                else:
                    YAECS_LOGGER.info(f"Pattern parameter '{key}' will be merged into the following matched "
                                      f"parameters : {list(to_merge.keys())}.")
            self.init_from_config(to_merge)
            return

        name = key.split('.')[0]
        attribute_name = "___" + name if name in self._methods else name
        try:
            old_value = getattr(self, attribute_name)
        except AttributeError as exception:
            raise AttributeError(f"ERROR : parameter '{key}' cannot be merged : '{name}' is not in the default "
                                 f"'{self.get_name().upper()}' config.\n{self._did_you_mean(key)}") from exception

        if "." in key:
            if isinstance(old_value, _ConfigurationBase):
                old_value.init_from_config({key.split('.', 1)[1]: value})
            else:
                did_you_mean = self._did_you_mean(name, filter_type=self.__class__, suffix=key.split('.', 1)[1])
                raise TypeError(f"Failed to set parameter '{key}' : '{name}' is not a sub-config.\n{did_you_mean}")

        else:
            if isinstance(old_value, _ConfigurationBase):
                if isinstance(value, _ConfigurationBase):
                    value = value.get_dict(deep=False, pre_post_processing_values=False)
                if not isinstance(value, dict):
                    raise TypeError(f"Trying to set sub-config '{old_value.get_name()}'\n"
                                    f"with non-config element '{value}'.\nThis replacement cannot be performed.")
                old_value.init_from_config(value)
            else:
                if isinstance(value, _ConfigurationBase):
                    value = value.get_dict(deep=True, pre_post_processing_values=False)
                self._set_parameter(name, attribute_name, value, old_value)

    def _add_item(self, key: str, value: Any) -> None:
        """ Method called by _process_item_to_merge_or_add if the value should be added and not merged (ie., only while
        setting up the default config). This method ultimately performs all additions to the config. """
        if "*" in key:
            raise ValueError(f"The '*' character is not authorised in the default config ({key}).")

        name = key.split('.')[0]
        attribute_name = "___" + name if name in self._methods else name

        if "." in key:
            try:
                sub_config = getattr(self, attribute_name)
            except AttributeError:
                sub_config = self._set_sub_config(name, attribute_name)
            if isinstance(sub_config, _ConfigurationBase):
                sub_config.init_from_config({key.split(".", 1)[1]: value})
            else:
                did_you_mean = self._did_you_mean(name, filter_type=self.__class__, suffix=key.split(".", 1)[1])
                raise TypeError(f"Failed to set parameter '{key}' : '{name}' is not a sub-config.\n{did_you_mean}")

        else:
            try:
                _ = getattr(self, attribute_name)
                raise RuntimeError(f"ERROR : parameter '{name}' was set twice.")
            except AttributeError:
                if isinstance(value, _ConfigurationBase):
                    self._set_sub_config(name, attribute_name, {k: value[k] for k in value.get_parameter_names(False)})
                else:
                    self._set_parameter(name, attribute_name, value)

    def _set_parameter(self, name: str, attribute_name: str, value: Any, old_value: Any = NoValue()) -> None:
        """ Method called by _add_item and _merge_item to set a parameter in the config to a new value. Ultimately
        performs the setting of all parameters in the config. """
        if name != attribute_name and isinstance(old_value, NoValue) and self._verbose:
            YAECS_LOGGER.warning(f"WARNING : '{name}' is the name of a method in the Configuration object.\n"
                                 f"Your parameter was initialised anyways, under the name {attribute_name}. You can "
                                 f"access it via config.{attribute_name} or config['{name}'].")
        if self._verbose:
            old_value_message = "" if isinstance(old_value, NoValue) else f"old : '{old_value}'\n"
            YAECS_LOGGER.debug(f"Setting '{name}' : \n{old_value_message}new : '{value}'.")

        full_name = self._get_full_path(name)
        self._main_config.remove_value_before_postprocessing(full_name)
        self.get_setter()(names={full_name: self._get_full_path(attribute_name)}, values={full_name: value},
                          processing_type="pre", container=self)
        if full_name not in self.get_modified_buffer():
            self.get_modified_buffer().append(full_name)

    def _set_sub_config(self, name: str, attribute_name: str, content: Optional[dict] = None) -> 'Configuration':
        """ Method called by _add_item to add a sub-config to a config. First an empty config is created, then its
        values are added with its init_from_config method. """
        sub_config = self._get_instance(
            name=name,
            overwriting_regime=(self._main_config.config_metadata["overwriting_regime"]),
            config_path_or_dictionary={} if content is None else content,
            state=self._state,
            nesting_hierarchy=self._nesting_hierarchy + [attribute_name],
            main_config=self._main_config,
            verbose=self._verbose
        )
        object.__setattr__(self, attribute_name, sub_config)
        return sub_config

    def _gather_command_line_dict(self, to_merge: Optional[Union[List[str], str]] = None) -> Dict[str, Any]:
        """ Method called automatically at the end of each constructor to gather all parameters from the command line
        into a dictionary. This dictionary is then merged. """

        if to_merge is not None:
            if isinstance(to_merge, list):
                to_merge = " ".join(to_merge)
            list_to_merge = get_quasi_bash_sys_argv(to_merge)
        else:
            list_to_merge = sys.argv

        # Setting the config to operational mode in case this is called manually
        object.__setattr__(self, "_operating_creation_or_merging", True)

        # Gather parameters, their values and their types
        to_merge = {}  # {param_name: new_value, ...}
        found_config_path = not bool(self._from_argv)
        in_param = []
        un_matched_params = []
        for element in list_to_merge:
            if element.startswith("--") and (found_config_path or element.split("=", 1)[0] != self._from_argv):
                if "=" in element:
                    pattern, value = element[2:].split("=", 1)
                    value = value if value != "" else None
                else:
                    pattern, value = element[2:], None
                in_param = []
                for parameter in self.get_parameter_names(deep=True):
                    if compare_string_pattern(parameter, pattern):
                        in_param.append(parameter)
                        to_merge[parameter] = value
                if not in_param:
                    un_matched_params.append(pattern)
            elif element.startswith("--"):
                in_param = []
                found_config_path = True
            elif in_param and to_merge[in_param[0]] is None:
                for parameter in in_param:
                    to_merge[parameter] = element
            elif in_param:
                for parameter in in_param:
                    to_merge[parameter] = f"{to_merge[parameter]} {element}"

        if un_matched_params and self._verbose:
            YAECS_LOGGER.warning(f"WARNING : parameters {un_matched_params}, encountered while merging params from the "
                                 f"command line, do not match any param in the config. They will not be merged.")

        # Infer types, then return
        return {key: yaml.safe_load("true" if val is None else val) for key, val in to_merge.items()}

    def _post_process_modified_parameters(self) -> None:
        """ This method is called at the end of a config creation or merging operation. It applies post-processing to
        all parameters modified by this operation. If a parameter is converted into a non-native YAML type, also keeps
        its former value in memory for saving purposes. """
        modified = [self.get_modified_buffer().pop(0) for _ in range(len(self.get_modified_buffer()))]
        splits = [name.split(".") for name in modified if name.startswith(".".join(self._nesting_hierarchy))]
        names = {".".join(s): ".".join(s[:-1] + ["___" + s[-1]]) if s[-1] in self._methods else ".".join(s)
                 for s in splits}
        values = {name: self._main_config[name] for name in names}

        self.get_setter()(names=names, values=values, processing_type="post", container=self,
                          only_set_processed_parameters=True)
        for name in names:
            try:
                should_save = values[name] != self._main_config[name]
            except Exception:
                should_save = True
            if should_save:
                self._main_config.save_value_before_postprocessing(name, copy.deepcopy(values[name]))

    def _parse_metadata(self, metadata: str) -> Tuple[str, str]:
        """ Parses metadata string to get the saving time, regime and variation name if there is one. """
        pattern = "Saving time : * (*) ; Regime : *"
        if not isinstance(metadata, str) or not compare_string_pattern(metadata, pattern):
            raise RuntimeError("'config_metadata' is a special parameter. Please do not edit or set it.")
        if metadata.count(";") == 1:
            time_chunk, regime_chunk = metadata.split(";")
            variation_chunk = None
        else:
            time_chunk, regime_chunk, variation_chunk = metadata.split(";")
        former_saving_time = float(time_chunk.split("(")[-1].split(")")[0].strip())
        regime = regime_chunk.split(":")[1].strip()
        if regime == "unsafe" and self._verbose:
            YAECS_LOGGER.warning("WARNING : YOU ARE LOADING AN UNSAFE CONFIG FILE. Reproducibility with "
                                 "corresponding experiment is not ensured.")
        elif regime not in ["auto-save", "locked"]:
            raise ValueError("'overwriting_regime' is a special parameter. It can only be set to 'auto-save'"
                             "(default), 'locked' or 'unsafe'.")
        variation = None if variation_chunk is None else variation_chunk.split(":")[1].strip()
        return former_saving_time, regime, variation

    def _scan_yaml_files(self, path: str) -> List[str]:
        """ Scans a YAML file for its parameters, gathers and adds type hints and processing functions. """
        path = self._find_path(path)
        scanner = YAMLScanner(path)

        if not any(state.startswith("setup") for state in self._state):
            ignored_hints = [param for param, type_hint in scanner.type_hints.items()
                             if not is_dict_type_hint(type_hint)]
            if ignored_hints and self._verbose:
                YAECS_LOGGER.warning("WARNING : type-hinting only has effect in the default config (except for dict "
                                     f"type hints). The type hints {ignored_hints} in file '{path}' will be ignored.")
            ignored_processors = list(scanner.processing_functions.keys())
            if ignored_processors and self._verbose:
                YAECS_LOGGER.warning("WARNING : registering processing functions only has effect in the default config."
                                     f" The functions {ignored_processors} in file '{path}' will be ignored.")
        registered_methods = self.get_setter().registered_methods.keys()
        processors_in_hints = {param: type_hint.split(",") for param, type_hint in scanner.type_hints.items()
                               if all(method in registered_methods for method in type_hint.split(","))}
        if processors_in_hints and self._verbose:
            YAECS_LOGGER.warning("WARNING : registering processing functions using !type:<method_name> is deprecated. "
                                 "It will still work until the next release, but you should switch to using "
                                 f"!<method_name> instead (detected in tags '{processors_in_hints}' in file '{path}').")

        type_hints = {param: type_hint for param, type_hint in scanner.type_hints.items()
                      if ((any(state.startswith("setup") for state in self._state) or is_dict_type_hint(type_hint))
                      and any(method not in registered_methods for method in type_hint.split(",")))}
        self.get_setter().bulk_add_type_hints({self._get_full_path(pattern): type_hint
                                               for pattern, type_hint in type_hints.items()}, source=path)

        processors = {param: methods for param, methods in scanner.processing_functions.items()
                      if any(state.startswith("setup") for state in self._state)
                      and all(method in registered_methods for method in methods)}
        processors = {**processors, **processors_in_hints}
        self.get_setter().bulk_add_processors({self._get_full_path(pattern): methods
                                               for pattern, methods in processors.items()},
                                              source=path, container=self)
        return scanner.params
